#
#    regfilterLib - routines to mess with registry dumps (regutils package)
#    Copyright (C) 1998 Memorial University of Newfoundland
#    
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.
#    
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#    
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#    

#
# Library out routines to read windows 9x registry dumps, to print dumps
# that have been read, to read `filter' files (used by regfilter) and to
# apply filters to dumps.
#

package regfilter;

use strict;
use IO::File;

use vars qw($VERSION);
$VERSION = '0.10';

use vars qw($errorString);

$errorString = undef;


#
# Read a filter config file: returns a ref to what is needed
# by readKey().
#
# Format of config file:
#	# Lines starting with # are comments
#
#	# Include the named file (dir includee checked first, then current dir)
#	include "somefile"
#
#	# Any keys that match the specified key pattern will have
#	# the following actions applied
#	[keyPattern]
#	    # Any entry within a matching key whose name matches namePattern
#	    # wlll have the following actions applied
#	    name namePattern
#		# Apply the given perl substitute command to the name field
#		subst name s/namePattern/perlPattern/perlSubstOptions
#		# Apply the given perl substitute command to the value field
#		subst value s/valuePattern/perlPattern/perlSubstOptions
#	    name namePattern
#		# delete the entry
#		delete
#	    # Any entry within a matching key whose value matches namePattern
#	    # wlll have the following actions applied
#	    value valuePattern
#		subst name s/namePattern/perlPattern/perlSubstOptions
#		subst value s/valuePattern/perlPattern/perlSubstOptions
#	    value namePattern
#		# delete the entry
#		delete
#	# Whole keys can be deleted using
#	[keyPattern] -
#
#
# Data structure created is (for internal use only):
#   [
#	{
#	    'keyPat' => "key\\name\pattern", # Note: leading HKEY_..\\ stripped
#	    'deleteAll' => 0 or 1,	# true if whole key to be deleted
#	    'keyActions' => [
#			{
#			  'matchOn' => 'name' or 'value',
#			  'matchPat' => "perlPattern",
#			  'entryActions' => [
#					{ 'action' => 'substName'
#							or 'substValue'
#							or 'delete',
#					  'subst' => perlSubstExpr,
#					},
#					...
#				]
#			}, ...
#		],
#	}, ...
#   ]
#
#
# This a IniFile::Filter are very (very) similar - keep them that way
{
    package RegFile::Filter;

    use strict;
    use Carp;

    use vars qw(&new &readConf &filter);
    use vars qw($errorString);

    sub new
    {
	my $proto = shift;
	confess "new RegFile::Filter takes no arguments" if @_ != 0;

	my $field = {
		    'keyFilters' => []
		};

	bless $field, (ref($proto) || $proto);
	return $field;
    }

    sub readConf
    {
	my($filter, $file) = @_;

	my($keyActions) = undef;
	my($entryActions) = undef;

	my $in = new IO::File "$file", "r";
	if (!defined $in) {
	    $errorString = "can't open $file - $!";
	    return 0;
	}

	while (<$in>) {
	    next if (/^\s*(#|$)/);
	    # Trim whitespace
	    s/^\s+//; s/\s+$//;

	    # include another filter file?
	    if (/^\s*include\s+"([^"]*)"\s*$/) {
		my($ifile) = ($1);

		# End the previous key (with error checking)
		if (defined $keyActions && !@$keyActions) {
		    $errorString = "$file:$.: previous key has no actions";
		    $in->close;
		    return 0;
		} elsif (defined $entryActions && !@$entryActions) {
		    $errorString = "$file:$.: previous name/value pattern has no actions";
		    $in->close;
		    return 0;
		}
		$keyActions = $entryActions = undef;

		# If not an absolute path, try relative to this file first.
		my($mypath) = $file;
		$mypath =~ s:/+[^/]*$::;
		if ($ifile !~ /^\// && $mypath ne '' && $mypath ne $file
		    && -e "$mypath/$ifile")
		{
		    return 0 if (!$filter->readConf("$mypath/$ifile"));
		} else {
		    return 0 if (!$filter->readConf($ifile));
		}
		next;
	    }

	    # A new key?
	    if (/^\[\w+\\\\(.+)](|\s*-)$/) {
		my($keyPat, $isDel) = ($1, $2 eq '' ? 0 : 1);

		if (defined $keyActions && !@$keyActions) {
		    $errorString = "$file:$.: previous key has no actions";
		    $in->close;
		    return 0;
		} elsif (defined $entryActions && !@$entryActions) {
		    $errorString = "$file:$.: previous name/value pattern has no actions";
		    $in->close;
		    return 0;
		}

		my $ka;
		push(@{$filter->{'keyFilters'}}, {
			'keyPat' => $1,
			'deleteAll' => $isDel,
			'keyActions' => ($ka = [ ])
		    });
		$keyActions = $isDel ? undef : $ka;
		next;
	    }
	    # A new entry match operator?
	    if (/^\s*(name|value)\s+(\S.*)$/) {
		my($matchOn, $matchPat) = ($1, $2);
		if (!defined $keyActions) {
		    $errorString = "$file:$.: name/value pattern found outside key";
		    $in->close;
		    return 0;
		}
		if (defined $entryActions && !@$entryActions) {
		    $errorString = "$file:$.: previous name/value pattern has no actions";
		    $in->close;
		    return 0;
		}
		$entryActions = [];
		push(@$keyActions, { 'matchOn' => $matchOn,
				    'matchPat' => $matchPat,
				    'entryActions' => $entryActions,
				    });
		next;
	    }
	    # A entry substitution
	    if (/^\s*subst\s+(name|value)\s+(\S.*)$/) {
		my($action, $subst) = ($1, $2);
		if (!defined $keyActions) {
		    $errorString = "$file:$.: substitution found outside key";
		    $in->close;
		    return 0;
		}
		if (!defined $entryActions) {
		    $errorString = "$file:$.: substitution found outside entry/key pattern";
		    $in->close;
		    return 0;
		}
		push(@$entryActions, { 'action' => "subst\u$action",
				       'subst' => $subst,
				       });
		next;
	    }
	    # A entry deletion
	    if (/^\s*delete$/) {
		if (!defined $keyActions) {
		    $errorString = "$file:$.: delete entry found outside key";
		    $in->close;
		    return 0;
		}
		if (!defined $entryActions) {
		    $errorString = "$file:$.: delete entry found outside entry/key pattern";
		    $in->close;
		    return 0;
		}
		push(@$entryActions, { 'action' => 'delete',
				       'subst' => undef,
				       });
		next;
	    }
	    $errorString = "$file:$.: unexpected line";
	    $in->close;
	    return 0;
	}
	if ($in->error()) {
	    $errorString = "$file: error reading - $!";
	    $in->close;
	    return 0;
	}
	$in->close;
	return 1;
    }

    #
    # Given a key info (as read by readKey) and a filter info (as read
    # by readFilterConf), modify the key info by filtering out the stuff
    # to be ignored and doing substitutions.  If everything is to be ignored,
    # the key info hash is emptied (%keyInfo == 0).
    #
    sub
    filter
    {
	my($filter, $keyInfo) = @_;

	if (!defined $keyInfo || !%$keyInfo) {
	    $errorString = "filter called with undefined or empty key info";
	    return undef;
	}

	my($key) = $keyInfo->{'key'};
	$key =~ s/^\w+\\//;	# strip leading key name...

	my $ign;
	foreach $ign (@{$filter->{'keyFilters'}}) {
	    next if ($key !~ /^$ign->{'keyPat'}$/i);
	    if ($ign->{'deleteAll'}) {
		foreach $key (keys %$keyInfo) {
		    delete $keyInfo->{$key};
		}
		return 1;	# no point in going on...
	    }
	    # Must be entry substituions/deletions
	    my $keyAction;
	    foreach $keyAction (@{$ign->{'keyActions'}}) {
		for (my $i = 0; $i < @{$keyInfo->{'entries'}}; $i++) {
		    my $target = $keyAction->{'matchOn'} eq 'name'
				    ? ${$keyInfo->{'entries'}}[$i]->{'name'}
				    : ${$keyInfo->{'entries'}}[$i]->{'value'};
		    $target =~ s/^"//; $target =~ s/"$//
			if ($keyAction->{'matchOn'} eq 'name');
		    next if ($target !~ /^$keyAction->{'matchPat'}$/i);
		    # Have a match - carry out entry actions
		    my $entryAction;
		    foreach $entryAction (@{$keyAction->{'entryActions'}}) {
			if ($entryAction->{'action'} eq 'substName') {
			    eval "\${\$keyInfo->{'entries'}}[\$i]->{'name'} =~ $entryAction->{'subst'}";
			    if ($@ ne '') {
				$errorString = "error substituting $target using $entryAction->{'subst'}";
				return undef;
			    }
			} elsif ($entryAction->{'action'} eq 'substValue') {
			    eval "\${\$keyInfo->{'entries'}}[\$i]->{'value'} =~ $entryAction->{'subst'}";
			    if ($@ ne '') {
				$errorString = "error substituting $target using $entryAction->{'subst'}";
				return undef;
			    }
			} elsif ($entryAction->{'action'} eq 'delete') {
			    splice(@{$keyInfo->{'entries'}}, $i, 1);
			    $i--;
			    last;
			} else {
			    $errorString = "regfilter::filter: internal error - unknown entry action: $entryAction->{'action'}";
			    return undef;
			}
		    }
		}
	    }
	}

	return 1;
    }
}


#
# Read a single key from the specified input.
# Returns undef on error, empty hash ref on eof,
# otherwise the following record is returned:
#	{
#	    'key' => "theName",		# name of key (with HKEY_...)
#	    'status' => 'd' | 'n' | 'e',# d: deleted, n:new, e:existing
#	    'entries' => [ {		# entries added/changed/deleted
#		    'name' => "\"theName\"",	# name of entry (maybe "@")
#		    'value' => "hex:...",	# value (undef if deleted)
#		    'oldValue' => "\"abc\"",	# undef if unknown (maybe "new")
#		} ]
#	}
#
sub
readKey
{
    my($fd) = @_;
    my($key, $stuff, $name, $value);
    my($info) = {};
    my($startLine, $line);
    my($firstLine) = 1;

    while (<$fd>) {
	$_ =~ s/\r$//;
	next if (/^REGEDIT4\s*$/ && $firstLine);
	if (/^\s*$/) {
	    last if (%$info);	# non empty key
	    next;
	}
	$firstLine = 0;
	if (/^\[(.*)](.*)/) {
	    if (%$info) {
		$errorString = "$.: keys not separated by blank line";
		return undef;
	    }
	    ($key, $stuff) = ($1, $2);
	    my($status);
	    if ($stuff =~ /^\s*$/) {
		$status = 'e';
	    } elsif ($stuff =~ /^\s*-\s*$/) {
		$status = 'd';
	    } elsif ($stuff =~ /^\s*\(new\)\s*$/) {
		$status = 'n';
	    } else {
		$errorString = "$.: unrecognized stuff after key: $stuff";
		return undef;
	    }
	    $info->{'key'} = $key;
	    $info->{'status'} = $status;
	    $info->{'entries'} = [];
	    next;
	}

	#
	# Must be a entry
	#
	if (!%$info) {
	    $errorString = "$.: entry found outside of key";
	    return undef;
	}
	#
	# Keep reading as long as we have an odd number of quotes...
	#
	$startLine = $.;
	my $sofar = '';
	while (1) {
	    if (/^(([^"\\]|\\\\|\\")+|"([^"\\]|\\\\|\\")*")*/) {
		$sofar .= $&;
		last if $' eq '';
		$_ = $';
	    }
	    $line = <$fd>;
	    $line =~ s/\r$//;
	    if (!defined $line) {
		$errorString = "$.: hit EOF looking for matching quote or non-continuation line (started at line $startLine)";
		return undef;
	    }
	    if ($_ eq "\\\n") {
		$_ = '';
		$line =~ s/^\s*//;
	    }
	    $_ .= $line;
	}
	$_ = $sofar;
	if (/^(\@|\"([^\\"]|\\"|\\\\)+")(|=(|(|stringz:|ustringz:)"([^\\"]|\\"|\\\\)*"|dword:[\da-fA-F]+|hex(|\([0-9a-fA-F]+\)):[\da-fA-F,]*))(|\s+\((.+)\))\s*$/)
	{
	    push(@{$info->{'entries'}}, {
		    'name' => $1,
		    'value' => $3 eq '' ? undef : $4,
		    'oldValue' => $8 eq '' ? undef : $9,
		}
	    );
	    next;
	}
	$errorString = "$.: can't extract entry name/info";
	return undef;
    }
    $errorString = undef;
    return $info;
}

#
# Print out a key, with all its info
#
sub
printKey
{
    my($out, $info) = @_;

    return if !defined $info || !%$info;

    print $out '[', $info->{'key'}, ']';
    print $out "\t(new)" if ($info->{'status'} eq 'n');
    print $out '-' if ($info->{'status'} eq 'd');
    print $out "\n";
    my $f;
    foreach $f (@{$info->{'entries'}}) {
	print $out $f->{'name'};
	print $out '=', $f->{'value'} if (defined $f->{'value'});
	print $out "\t(", $f->{'oldValue'}, ')' if (defined $f->{'oldValue'});
	print $out "\n";
    }
    print $out "\n";
}

#
# Compare two keys (case insensitive, handles \\ correctly)
#
sub
cmpKey
{
    my($x, $y) = @_;
    $x =~ tr/A-Z/a-z/;
    $x =~ s/\\/\001/g;	# \001 instead of \0 due to perl 5.003 bug
    $y =~ tr/A-Z/a-z/;
    $y =~ s/\\/\001/g;
    return $x cmp $y;
}

#
# Compare two entry names (case insensitive, handles @, strips quotes)
#
sub
cmpEntryName
{
    my($x, $y) = @_;

    if ($x eq '@') {
	$x = '';
    } else {
	$x =~ s/^"//;
	$x =~ s/"$//;
	$x =~ tr/A-Z/a-z/;
    }

    if ($y eq '@') {
	$y = '';
    } else {
	$y =~ s/^"//;
	$y =~ s/"$//;
	$y =~ tr/A-Z/a-z/;
    }

    return $x cmp $y;
}

1;
